@@ -29,6 +29,9 @@
 #ifdef HAVE_ERRNO_H
 #include <errno.h> /* for the ENOENT definitions in str_errors */
 #endif
+#ifdef HAVE_LIMITS_H
+#include <limits.h>
+#endif
 
 #include "ne_string.h"
 #include "ne_utils.h"
@@ -750,6 +753,41 @@ static int strhash_sha_512_256(void)
     return OK;
 }
 
+/* FIPS 180-1 standard SHA-1 test vectors from RFC 3174 */
+#define TEST1_SHA1 "abc"
+#define TEST1_SHA1_MD "a9993e364706816aba3e25717850c26c9cd0d89d"
+#define TEST2_SHA1 "abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq"
+#define TEST2_SHA1_MD "84983e441c3bd26ebaae4aa1f95129e5e54670f1"
+#define TEST2_SHA1_MDC "84:98:3e:44:1c:3b:d2:6e:ba:ae:4a:a1:f9:51:29:e5:e5:46:70:f1"
+
+static int strhash_sha1(void)
+{
+    char *p = ne_strhash(NE_HASH_SHA1, "", NULL);
+
+    if (p == NULL) {
+        t_context("SHA-1 not supported");
+        return SKIP;
+    }
+    ne_free(p);
+
+    /* Test vector 1: "abc" */
+    ONVEC((NE_HASH_SHA1, TEST1_SHA1, NULL), TEST1_SHA1_MD);
+
+    /* Test vector 2: "abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq" */
+    ONVEC((NE_HASH_SHA1, TEST2_SHA1, NULL), TEST2_SHA1_MD);
+
+    /* Test with colon formatting */
+    ONVEC((NE_HASH_SHA1|NE_HASH_COLON, TEST2_SHA1, NULL), TEST2_SHA1_MDC);
+
+    /* Test with multiple string arguments */
+    ONVEC((NE_HASH_SHA1, "a", "b", "c", NULL), TEST1_SHA1_MD);
+
+    /* Test empty string - da39a3ee5e6b4b0d3255bfef95601890afd80709 */
+    ONVEC((NE_HASH_SHA1, "", NULL), "da39a3ee5e6b4b0d3255bfef95601890afd80709");
+
+    return OK;
+}
+
 static int strparam(void)
 {
     static const struct {
@@ -758,6 +796,7 @@ static int strparam(void)
         const char *expect;
     } ts[] = {
         { "UTF-8", NULL, "foobar", NULL },
+        { "UTF-8", NULL, "foo12345bar", NULL },
         { "UTF-8", NULL, "foo@bar", "UTF-8''foo%40bar" },
         { "UTF-8", NULL, "foo bar", "UTF-8''foo%20bar" },
         { "iso-8859-1", "en", "\xA3 rates", "iso-8859-1'en'%a3%20rates" },
@@ -770,7 +809,8 @@ static int strparam(void)
         char *act = ne_strparam(ts[n].charset, ts[n].lang, (const unsigned char *)ts[n].value);
 
         if (ts[n].expect == NULL) {
-            ONV(act != NULL, ("expected NULL output for '%s'", ts[n].value));
+            ONV(act != NULL, ("expected NULL output for '%s', got '%s'",
+                              ts[n].value, act));
         }
         else {
             ONCMP(act, ts[n].expect);
@@ -781,6 +821,145 @@ static int strparam(void)
     return OK;
 }
 
+static int strhextoul(void)
+{
+    static const struct {
+        const char *input;
+        unsigned long expect;
+        int error; /* errno */
+        const char *endp;
+    } *t, ts[] = {
+        { "0x0", ULONG_MAX, EINVAL },
+        { "0X1", ULONG_MAX, EINVAL },
+        { "+1", ULONG_MAX, EINVAL },
+        { "-0", ULONG_MAX, EINVAL },
+        { "+0x1", ULONG_MAX, EINVAL },
+        { "+0X1", ULONG_MAX, EINVAL },
+        { "", ULONG_MAX, EINVAL },
+        { "1", 1, 0, "" },
+        { " 10", ULONG_MAX, EINVAL, " 10" },
+        { "0000010 ", 16, 0, " " },
+        { "4242", 16962, 0 },
+        { "4242zZZz", 16962, 0, "zZZz" },
+        { "cAfEBeEf", 3405692655 },
+#if SIZEOF_LONG == 8
+        { "10000000000000000", ULONG_MAX, ERANGE },
+        { "100000000", 0x100000000, 0 },
+#elif SIZEOF_LONG == 4
+        { "100000000", ULONG_MAX, ERANGE },
+#endif
+        { NULL }
+    };
+    unsigned n;
+
+    for (n = 0; ts[n].input; n++) {
+        unsigned long actual;
+        const char *endp = "(unset)", **endpp;
+        int errnum;
+
+        t = ts + n;
+
+        endpp = t->endp ? &endp : NULL;
+        errno = ENOENT;
+        actual = ne_strhextoul(t->input, endpp);
+        errnum = errno;
+        ONV(errnum != t->error,
+            ("got errno %d not %d for [%s]", errnum, t->error, t->input));
+        ONV(actual != t->expect,
+            ("got %lu not %lu for [%s]", actual, t->expect, t->input));
+        if (endpp) ONCMP(*endpp, t->endp);
+    }
+
+    return OK;
+}
+
+static int strlower(void)
+{
+    char buf[256];
+
+    ne_strnzcpy(buf, "HELLO", sizeof buf);
+    ONN("unexpected return value", ne_strlower(buf) != buf);
+    ONV(strcmp(buf, "hello"), ("strlower failed: got %s", buf));
+
+    ne_strnzcpy(buf, "MiXeD CaSe 123", sizeof buf);
+    ne_strlower(buf);
+    ONV(strcmp(buf, "mixed case 123"), ("strlower failed: got %s", buf));
+
+    ne_strnzcpy(buf, "already lower", sizeof buf);
+    ne_strlower(buf);
+    ONV(strcmp(buf, "already lower"), ("strlower failed: got %s", buf));
+
+    ne_strnzcpy(buf, "", sizeof buf);
+    ne_strlower(buf);
+    ONV(strcmp(buf, ""), ("strlower failed on empty string"));
+
+    return OK;
+}
+
+static int strupper(void)
+{
+    static const struct {
+        const char *input;
+        const char *expected;
+    } ts[] = {
+        { "hello", "HELLO" },
+        { "MiXeD CaSe 123", "MIXED CASE 123" },
+        { "ALREADY UPPER", "ALREADY UPPER" },
+        { "", "" },
+        { NULL, NULL }
+    };
+    unsigned int i;
+
+    for (i = 0; ts[i].input != NULL; i++) {
+        char buf[256];
+
+        ne_strnzcpy(buf, ts[i].input, sizeof buf);
+        
+        ONN("unexpected return value", ne_strupper(buf) != buf);
+        ONV(strcmp(buf, ts[i].expected),
+            ("strupper failed: got %s, expected %s", buf, ts[i].expected));
+    }
+
+    return OK;
+}
+
+static int mknonce(void)
+{
+    size_t ts[3] = {16, 32, 64};
+    unsigned int n;
+
+    for (n = 0; n < sizeof(ts)/sizeof(ts[0]); n++) {
+        unsigned char nonce1[256], nonce2[256];
+
+        ONN("nonce generation failed", ne_mknonce(nonce1, ts[n], 0));
+        ONN("second nonce generation failed", ne_mknonce(nonce2, ts[n], 0));
+        ONV(memcmp(nonce1, nonce2, ts[0]) == 0,
+            ("two consecutive nonces are identical"));
+    }
+
+    return OK;
+}
+
+/* Test multiple nonces for uniqueness */
+static int mknonce_uniqueness(void)
+{
+    unsigned char nonces[10][16];
+    unsigned int i, j;
+
+    /* Generate 10 nonces */
+    for (i = 0; i < 10; i++)
+        ONN("nonce generation failed", ne_mknonce(nonces[i], sizeof nonces[i], 0));
+
+    /* Check all are unique */
+    for (i = 0; i < 10; i++)
+        for (j = i + 1; j < 10; j++)
+            ONV(memcmp(nonces[i], nonces[j], sizeof nonces[i]) == 0,
+                ("nonces at positions %u and %u are identical",
+                 i, j));
+
+    return OK;
+}
+
 ne_test tests[] = {
     T(simple),
     T(buf_concat),
@@ -810,10 +989,16 @@ ne_test tests[] = {
     T(buf_print),
     T(qappend),
     T(strhash),
+    T(strhash_sha1),
     T(strhash_sha_256),
     T(strhash_sha_512),
     T(strhash_sha_512_256),
     T(strparam),
+    T(strhextoul),
+    T(strlower),
+    T(strupper),
+    T(mknonce),
+    T(mknonce_uniqueness),
     T(NULL)
 };
 
